use std::{
    fmt::Display,
    io::{Error as IOError, ErrorKind},
    path::StripPrefixError,
    string::FromUtf8Error,
};

#[derive(Debug)]
pub struct Error {
    message: String,
    source: Option<Box<dyn std::error::Error>>,
}

impl Error {
    pub fn new<E: std::error::Error + 'static>(message: String, source: E) -> Self {
        Self {
            message,
            source: Some(Box::new(source)),
        }
    }

    fn wrap<E: std::error::Error + 'static>(value: E) -> Option<Box<dyn std::error::Error>> {
        Some(Box::new(value))
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", &self.message)?;

        if let Some(source) = &self.source {
            write!(f, " ({})", &source)?;
        }

        Ok(())
    }
}

impl std::error::Error for Error {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        self.source.as_deref()
    }
}

impl From<String> for Error {
    fn from(value: String) -> Self {
        Self {
            message: value,
            source: None,
        }
    }
}

impl From<StripPrefixError> for Error {
    fn from(value: StripPrefixError) -> Self {
        Self {
            message: value.to_string(),
            source: Self::wrap(value),
        }
    }
}

impl From<IOError> for Error {
    fn from(value: IOError) -> Self {
        Self {
            message: value.to_string(),
            source: Self::wrap(value),
        }
    }
}

impl From<Error> for IOError {
    fn from(value: Error) -> Self {
        IOError::new(ErrorKind::Other, value.message)
    }
}

impl From<FromUtf8Error> for Error {
    fn from(value: FromUtf8Error) -> Self {
        Self {
            message: value.to_string(),
            source: Self::wrap(value),
        }
    }
}

impl From<&'static dyn std::error::Error> for Error {
    fn from(value: &'static dyn std::error::Error) -> Self {
        Self {
            message: value.to_string(),
            source: Self::wrap(value),
        }
    }
}
